/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#ifndef _WIN32
#define _GNU_SOURCE
#endif
/*-----------------------------------------------------------------------------*
 // include
 *-----------------------------------------------------------------------------*/
#include <unistd.h>
#include <errno.h>
#include <libxml/xpath.h>
#include "harktool.h"

/*-----------------------------------------------------------------------------*
 // extern
 *-----------------------------------------------------------------------------*/
extern struct GLOBAL g;

/*-----------------------------------------------------------------------------*
 * 新規に配列を作成(main)
 *
 *
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_new(int type) {

    switch (type) {
    //伝達関数用３次元配列を作成する(complex)
    //Nsrc x Nmic x Nfft/2+1
    case TF_VECT:
        g.TFvect = harktool_newComplexFloatThreeDim(g.Nsrc, g.Nmic, (g.Nfft / 2 + 1));
        break;

        //分離伝達関数用３次元配列を作成する(complex)
        //Nsrc x Nmic x Nfft/2+1
    case SEP_TF_VECT:
        g.SepTFvect = harktool_newComplexFloatThreeDim(g.Nsrc, g.Nmic, (g.Nfft / 2 + 1));
        break;

        //伝達関数作業用３次元配列を作成する(complex)
        //*  Nsrc x Nmic x Nfft/2+1
    case TMP_TF_VECT:
        g.TmpTFvect = harktool_newComplexFloatThreeDim(g.Nsrc, g.Nmic, g.Nfft / 2 + 1);
        break;

    case IR_VECT:
        //３次元配列を作成する(float) // Nsrc x Nmic x Nfft
        g.LocIRvect = harktool_newFloatThreeDim(g.Nsrc, g.Nmic, g.Nfft);
        g.SepIRvect = harktool_newFloatThreeDim(g.Nsrc, g.Nmic, g.Nfft);
        break;

    case TMP_IR_VECT:
        //* impulse応答用領域を作成する //Nsrc x Nmic x Nfft
        g.TmpIRvect = harktool_newFloatThreeDim(g.Nsrc, g.Nmic, g.Nfft);
        break;

        //インパルス応答(waveファイル)格納用２次元配列を作成する(float)
    case WAV_DATA:
//        g.wav.data = harktool_newFloatTwoDim(  g.Nmic , g.Nfft );
        break;

        //mic position格納用２次元配列を作成する(float) // Nmic * 3 (x,y,z)
    case MIC_POS_VECT:
        g.MicPosVect = harktool_newFloatTwoDim(g.Nmic, 3);
        break;

    case SRC_POS_VECT:
        g.SrcPosVect = harktool_newFloatTwoDim(g.Nsrc, 3);
        break;

        //CalcSysimpで作成したImpulse応答ファイルパス設定領域を作成する (2次元配列)
        //* TSPファイル名格納領域を作成する (2次元配列)
    case FLT_PATH:
        g.tmp.flt_path = harktool_newCharTwoDim(g.Nsrc, PATH_MAX);
        break;

    case TSP_WAV_NAME:
        g.tmp.tsp_wav_name = harktool_newCharTwoDim(g.Nsrc, PATH_MAX);
        break;

    case IR_WAV_NAME:
//      harktool_newCharTwoDim( g.tmp.  .ir_wav_path , g.Nsrc , PATH_LEN );
        break;

    case SORT_POS_ID:
        g.MicPosId = harktool_newIntOneDim(g.Nmic);
        g.SrcPosId = harktool_newIntOneDim(g.Nsrc);
        break;

    case BUFFER:
        g.buff_a = harktool_newComplexFloatOneDim(g.Nfft);
        g.buff_A = harktool_newComplexFloatOneDim(g.Nfft);
        break;

    }

}        //end

/*-----------------------------------------------------------------------------*
 * 配列をfreeする
 *
 *
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_delete(int type) {

    switch (type) {
    //伝達関数用３次元配列を解放する(complex)
    case TF_VECT:
        harktool_freeComplexFloatThreeDim(g.TFvect);
        g.TFvect = NULL;
        break;

        //伝達関数作業用３次元配列を解放する(complex)
        //*  Nsrc x Nmic x Nfft/2+1
    case TMP_TF_VECT:
        harktool_freeComplexFloatThreeDim(g.TmpTFvect);
        g.TmpTFvect = NULL;
        break;

    case IR_VECT:
        //３次元配列を解放する(float) // Nsrc x Nmic x Nfft
        harktool_freeFloatThreeDim(g.LocIRvect);
        g.LocIRvect = NULL;
        harktool_freeFloatThreeDim(g.SepIRvect);
        g.SepIRvect = NULL;
        break;

    case TMP_IR_VECT:
        //* impulse応答用領域を解放する //Nsrc x Nmic x Nfft
        harktool_freeFloatThreeDim(g.TmpIRvect);
        g.TmpIRvect = NULL;
        break;

        //インパルス応答(waveファイル)格納用２次元配列を解放する(float)
    case WAV_DATA:
        harktool_freeFloatTwoDim(g.wav.data);
        g.wav.data = NULL;
        break;

        //mic position格納用２次元配列を解放する(float) // Nmic * 3 (x,y,z)
    case MIC_POS_VECT:
        harktool_freeFloatTwoDim(g.MicPosVect);
        g.MicPosVect = NULL;
        break;

        //CalcSysimpで作成したImpulse応答ファイルパス設定領域を解放する (2次元配列)
        //* TSPファイル名格納領域を解放する (2次元配列)
    case FLT_PATH:
        harktool_freeCharTwoDim(g.tmp.flt_path);
        g.tmp.flt_path = NULL;
        break;

    case TSP_WAV_NAME:
        harktool_freeCharTwoDim(g.tmp.tsp_wav_name);
        g.tmp.tsp_wav_name = NULL;
        break;

    case IR_WAV_NAME:
//      harktool_freeCharTwoDim( tmp.ir_wav_path , g.Nsrc  );
        break;

    case SORT_POS_ID:
        harktool_freeIntOneDim(g.MicPosId);
        harktool_freeIntOneDim(g.SrcPosId);
        g.MicPosId = NULL;
        g.SrcPosId = NULL;
        break;

    case BUFFER:
        harktool_freeComplexFloatOneDim(g.buff_a);
        harktool_freeComplexFloatOneDim(g.buff_A);
        g.buff_a = NULL;
        g.buff_A = NULL;
        break;

    }

}        //

/*-----------------------------------------------------------------------------*
 * 1次元配列
 *-----------------------------------------------------------------------------*/
#define newOneDim(type, size_x) {\
    type* array = (type*) malloc(size_x * sizeof(type));\
    if (array == NULL) {\
        harktool_LogE("can not allocated memory");\
        harktool_Exit(EXIT_FAILURE);\
    }\
    return array;\
}

#define freeOneDim(array) {\
    if (array != NULL) {\
        free(array);\
    }\
}

// int
int* harktool_newIntOneDim(int size_x) {
    newOneDim(int, size_x);
}
void harktool_freeIntOneDim(int* array) {
    freeOneDim(array);
}

// COMPLEX_FLOAT
COMPLEX_FLOAT* harktool_newComplexFloatOneDim(int size_x) {
    newOneDim(COMPLEX_FLOAT, size_x);
}
void harktool_freeComplexFloatOneDim(COMPLEX_FLOAT* array) {
    freeOneDim(array);
}

/*-----------------------------------------------------------------------------*
 * 2次元配列
 *-----------------------------------------------------------------------------*/
#define newTwoDim(type, size_x, size_y) {\
    type** matrix = (type**)malloc(sizeof(type*) * size_x);\
    matrix[0] = (type*)calloc(size_x * size_y, sizeof(type));\
    if (!matrix || !matrix[0]) {\
        harktool_LogE("can not allocated memory");\
        harktool_Exit(EXIT_FAILURE);\
    }\
    for (int i = 1; i < size_x; i++) {\
        matrix[i] = matrix[i - 1] + size_y;\
    }\
    return matrix;\
}

#define freeTwoDim(martix) {\
    if (matrix) {\
        if (matrix[0]) {\
            free(matrix[0]);\
        }\
        free(matrix);\
    }\
}

// char
char** harktool_newCharTwoDim(int size_x, int size_y) {
    newTwoDim(char, size_x, size_y);
}
void harktool_freeCharTwoDim(char **matrix) {
    freeTwoDim(matrix);
}

// float
float** harktool_newFloatTwoDim(int size_x, int size_y) {
    newTwoDim(float, size_x, size_y);
}
void harktool_freeFloatTwoDim(float** matrix) {
    freeTwoDim(matrix);
}

// COMPLEX_DOUBLE
COMPLEX_DOUBLE** harktool_newComplexDoubleTwoDim(int size_x, int size_y) {
    newTwoDim(COMPLEX_DOUBLE, size_x, size_y);
}
void harktool_freeComplexDoubleTwoDim(COMPLEX_DOUBLE** matrix) {
    freeTwoDim(matrix);
}

/*-----------------------------------------------------------------------------*
 * 3次元配列
 *-----------------------------------------------------------------------------*/
#define newThreeDim(type, size_x, size_y, size_z) {\
    type*** matrix = (type***) malloc(sizeof(type**) * size_x);\
    matrix[0] = (type**) malloc(sizeof(type**) * size_x * size_y);\
    matrix[0][0] = (type*) calloc(size_x * size_y * size_z, sizeof(type));\
    if (!matrix || !matrix[0] || !matrix[0][0]) {\
        harktool_LogE("can not allocated memory");\
        harktool_Exit(EXIT_FAILURE);\
    }\
    for (int i = 0; i < size_x; i++) {\
        matrix[i] = matrix[0] + i * size_y;\
        for (int j = 0; j < size_y; j++) {\
            matrix[i][j] = matrix[0][0] + (i * size_y + j) * size_z;\
        }\
    }\
    return matrix;\
}

#define freeThreeDim(matrix) {\
    if (matrix) {\
        if (matrix[0]) {\
            if (matrix[0][0]) {\
                free(matrix[0][0]);\
            }\
            free(matrix[0]);\
        }\
        free(matrix);\
    }\
}

// float
float*** harktool_newFloatThreeDim(int size_x, int size_y, int size_z) {
    newThreeDim(float, size_x, size_y, size_z);
}
void harktool_freeFloatThreeDim(float*** matrix) {
    freeThreeDim(matrix);
}

// COMPLEX_FLOAT
COMPLEX_FLOAT*** harktool_newComplexFloatThreeDim(int size_x, int size_y, int size_z) {
    newThreeDim(COMPLEX_FLOAT, size_x, size_y, size_z);
}
void harktool_freeComplexFloatThreeDim(COMPLEX_FLOAT*** matrix) {
    freeThreeDim(matrix);
}

/*-----------------------------------------------------------------------------*
 * ３次元配列データを初期化する(complex)
 *
 * Nsrc x Nmic x Nfft/2+1
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_initArray(int type) {
    switch (type) {
    case TMP_IR_VECT:
        for (int src = 0; src < g.Nsrc; src++) {
            for (int mic = 0; mic < g.Nmic; mic++) {
                memset(g.TmpIRvect[src][mic], '\0', sizeof(float) * g.Nfft);
            }
        }
        break;

    case TMP_TF_VECT:
        for (int src = 0; src < g.Nsrc; src++) {
            for (int mic = 0; mic < g.Nmic; mic++) {
                memset(g.TmpTFvect[src][mic], '\0', sizeof(COMPLEX_FLOAT) * (g.Nfft / 2 + 1));
            }
        }
        break;
    }
}

void getIrRange4FFT(enum TFType type, int farthest, int margin, int limit, int src, POSITION_INFO* posInfo) {
    char* strType = (type == LOC) ? "LOC" : "SEP";
    int start = farthest + margin - g.Nfft;
    if (start < 0)
        start = 0;

    int end = start + g.Nfft;
    if (end > limit)
        end = limit;

    posInfo->tfOffset = start;
    harkio_Log_printf(harkio_Log_Debug, " %s offset=%d(%fsec), extract %d - %d", strType, start,
            (float )start / (float )((g.SamplingRate > 0) ? g.SamplingRate : 16000.0), start, end);
    posInfo->cutFrom = start;
    posInfo->cutTo = end;
    if (start > 0) {
        harkio_Log_printf(harkio_Log_Info, "Src[%d] %s Cut Index %d - %d", g.SrcPosId[src], strType, start, end);
    }
}

void setIr4FFT(enum TFType type, int peak, const POSITION_INFO* posInfo, int src, int mic, int soundLength,
        const float* srcIRvect, float* destIRvect) {
    int deltaPeak = peak - posInfo->cutFrom;
    if (deltaPeak < 0) {
        harkio_Log_printf(harkio_Log_Warning,
                "The direct sound data may have lost on src[%d], mic[%d]. Check the data or %s-length(%d).",
                g.SrcPosId[src], g.MicPosId[mic], (type == LOC) ? "direct" : "reverb", soundLength);
        cat_asprintf(&g.readmeMsg,
                "The direct sound data may have lost on src[%d], mic[%d]. Check the data or %s-length(%d).\n",
                g.SrcPosId[src], g.MicPosId[mic], (type == LOC) ? "direct" : "reverb", soundLength);
    }
    memcpy(destIRvect, &srcIRvect[posInfo->cutFrom], sizeof(float) * (posInfo->cutTo - posInfo->cutFrom));
    if (type == LOC) {
        // 残響音と思われる部分に0を書き込む
        for (int i = deltaPeak + 1 + soundLength; i < g.Nfft; i++) {
            destIRvect[i] = 0.0f;
        }
    }
}

void harktool_setIRvectFromFlt(enum TFType outputType, int src, int directLength, int reverbLength,
        SRC_POS_INFO* srcPosInfo) {
    float* fltBufs[g.Nmic];
    int frameLengths[g.Nmic];
    int peaks[g.Nmic];
    for (int mic = 0; mic < g.Nmic; mic++) {
        char* fname = NULL;
        asprintf(&fname, "%s/float_ch%d.flt", g.tmp.flt_path[src], mic); // float_ch0.flt

        // 1flt file size 取得(16384byte)
        long file_size = 0;
        harktool_getFileSize(fname, &file_size);

        int fltBufLength = (file_size > INT_MAX) ? INT_MAX : file_size;
        frameLengths[mic] = fltBufLength / sizeof(float);

        // 1flt file size 取得(16384byte)
        fltBufs[mic] = (float *) malloc(fltBufLength);

        if (fltBufs[mic] == NULL) {
            harktool_LogE("can not allocated memory");
            harktool_Exit(EXIT_FAILURE);
        }

        //1ch分の file のデータをbufferに格納する
        harktool_SetFloatArray(fltBufs[mic], fltBufLength, fname);

        // peak search
        peaks[mic] = searchPeak(fltBufs[mic], frameLengths[mic], g.PeakSearchFrom - 1, g.PeakSearchTo - 1);

        if (fname)
            free(fname);
    }

    // calc start-end position to extract
    int nearest = peaks[0];
    int farthest = peaks[0];
    int minLength = frameLengths[0];
    for (int mic = 1; mic < g.Nmic; mic++) {
        if (nearest > peaks[mic])
            nearest = peaks[mic];
        if (farthest < peaks[mic])
            farthest = peaks[mic];
        if (minLength > frameLengths[mic])
            minLength = frameLengths[mic];
    }
    harkio_Log_printf(harkio_Log_Debug, "FLT IR[%d] peak=%d to %d", g.SrcPosId[src], nearest, farthest);

    // ピーク位置の差が一定数以上ある場合、ピークは反射音の可能性がある
    showReverbSources(src, nearest, peaks);

    // calculate offset and cut index
    switch (outputType) {
    case LOC:
        getIrRange4FFT(LOC, farthest, directLength + 1, minLength, src, &srcPosInfo->loc);
        break;
    case SEP:
        getIrRange4FFT(SEP, farthest, reverbLength + 1, minLength, src, &srcPosInfo->sep);
        break;
    case LOC_SEP:
        getIrRange4FFT(LOC, farthest, directLength + 1, minLength, src, &srcPosInfo->loc);
        getIrRange4FFT(SEP, farthest, reverbLength + 1, minLength, src, &srcPosInfo->sep);
        break;
    }

    // extract IR to IRvect
    for (int mic = 0; mic < g.Nmic; mic++) {
        switch (outputType) {
        case LOC:
            setIr4FFT(LOC, peaks[mic], &srcPosInfo->loc, src, mic, directLength, fltBufs[mic], g.LocIRvect[src][mic]);
            break;
        case SEP:
            setIr4FFT(SEP, peaks[mic], &srcPosInfo->sep, src, mic, reverbLength, fltBufs[mic], g.SepIRvect[src][mic]);
            break;
        case LOC_SEP:
            setIr4FFT(LOC, peaks[mic], &srcPosInfo->loc, src, mic, directLength, fltBufs[mic], g.LocIRvect[src][mic]);
            setIr4FFT(SEP, peaks[mic], &srcPosInfo->sep, src, mic, reverbLength, fltBufs[mic], g.SepIRvect[src][mic]);
            break;
        }
        free(fltBufs[mic]);
    }
}

/*-----------------------------------------------------------------------------*
 * Impulse応答のwav fileを読み込んで配列に格納する
 * in  : wave.data[mic][fft];
 * out : param.IRvector
 *-----------------------------------------------------------------------------*/
void harktool_setIRvectFromWav(enum TFType outputType, int directLength, int reverbLength, SRC_POS_INFO srcPosInfos[]) {
    for (int src = 0; src < g.Nsrc; src++) {
        printf(" progress:%d / %d\n", src + 1, g.Nsrc);
        fflush(stdout);

        //mic_idのpositionを取得
        harkio_Position* pos = harkio_Positions_getById(g.src_xml->poses, g.SrcPosId[src]);

        int frameLength, samplingRate = 16000;
        // file のデータをbuffer(g.wav.data[ch][])に格納する
        harktool_getWavDataFromFile(pos->path, &frameLength, &samplingRate);

        int peaks[g.Nmic];
        for (int mic = 0; mic < g.Nmic; mic++) {
            peaks[mic] = searchPeak(g.wav.data[mic], frameLength, g.PeakSearchFrom - 1, g.PeakSearchTo - 1);
        }

        // calc start-end position to extract
        int nearest = peaks[0];
        int farthest = peaks[0];
        for (int mic = 1; mic < g.Nmic; mic++) {
            if (nearest > peaks[mic])
                nearest = peaks[mic];
            if (farthest < peaks[mic])
                farthest = peaks[mic];
        }
        harkio_Log_printf(harkio_Log_Debug, "WAV IR[%d] peak=%d to %d", g.SrcPosId[src], nearest, farthest);

        // ピーク位置の差が一定数以上ある場合、ピークは反射音の可能性がある
        showReverbSources(src, nearest, peaks);

        // calculate offset and cut index
        switch (outputType) {
        case LOC:
            getIrRange4FFT(LOC, farthest, directLength + 1, frameLength, src, &srcPosInfos[src].loc);
            break;
        case SEP:
            getIrRange4FFT(SEP, farthest, reverbLength + 1, frameLength, src, &srcPosInfos[src].sep);
            break;
        case LOC_SEP:
            getIrRange4FFT(LOC, farthest, directLength + 1, frameLength, src, &srcPosInfos[src].loc);
            getIrRange4FFT(SEP, farthest, reverbLength + 1, frameLength, src, &srcPosInfos[src].sep);
            break;
        }

        // extract IR to IRvect
        for (int mic = 0; mic < g.Nmic; mic++) {
            switch (outputType) {
            case LOC:
                setIr4FFT(LOC, peaks[mic], &srcPosInfos[src].loc, src, mic, directLength, g.wav.data[mic],
                        g.LocIRvect[src][mic]);
                break;
            case SEP:
                setIr4FFT(SEP, peaks[mic], &srcPosInfos[src].sep, src, mic, reverbLength, g.wav.data[mic],
                        g.SepIRvect[src][mic]);
                break;
            case LOC_SEP:
                setIr4FFT(LOC, peaks[mic], &srcPosInfos[src].loc, src, mic, directLength, g.wav.data[mic],
                        g.LocIRvect[src][mic]);
                setIr4FFT(SEP, peaks[mic], &srcPosInfos[src].sep, src, mic, reverbLength, g.wav.data[mic],
                        g.SepIRvect[src][mic]);
                break;
            }
        }
    }
}

// 指定されたフレーム数、チャンネルだけ読み込む
void harktool_getWavDataFromFile(const char *fname, int* frameLength, int* samplingRate) {

    SNDFILE *fp;    // ファイルハンドラ
    SF_INFO sfinfo; // チャンネル数、サンプリング周波数などを持つ構造体

    // sf_open の前に 0 にしておく
    sfinfo.format = 0;

    if ((fp = sf_open(fname, SFM_READ, &sfinfo)) == NULL) {
        harktool_LogE("can not open wave file [%s]", fname);
        harktool_Exit(EXIT_FAILURE);
    }

    *samplingRate = sfinfo.samplerate;

    //check channel
    for (int i = 0; i < g.Nmic; i++) {
        if (g.MicPosId[i] > sfinfo.channels) {
            sf_close(fp);
            harkio_Log_printf(harkio_Log_Error, "impulse data not found: wav channel [%d] in file [%s]", g.MicPosId[i],
                    fname);
            harktool_Exit(EXIT_FAILURE);
        }
    }

    // 読み込む frame数
    sf_count_t ndata = 1; //1frame分
    *frameLength = (sfinfo.frames > INT_MAX) ? INT_MAX : sfinfo.frames;

    float* in_buf = (float*) malloc(sizeof(float) * sfinfo.channels);

    if (g.wav.data == NULL) {
        g.wav.data = harktool_newFloatTwoDim(g.Nmic, *frameLength);
    }

    for (int frame = 0; frame < *frameLength; frame++) {
        int read_num = sf_readf_float(fp, in_buf, ndata); //1frame分

        if (read_num != ndata) {
            harkio_Log_printf(harkio_Log_Error, "read count different. read_num=%d count=%d", read_num, (int ) ndata);
            free(in_buf);
            sf_close(fp);
            harktool_Exit(EXIT_FAILURE);
        }

        for (int mic = 0; mic < g.Nmic; mic++) {
            g.wav.data[mic][frame] = in_buf[g.MicPosId[mic]];
        }
    }

    sf_close(fp);
    free(in_buf);

}

/*-----------------------------------------------------------------------------*
 * ユニークなディレクトリ名作成用に現在日時分秒を取得
 *
 *
 *  OUT : param.date_time    //取得した時間の文字列
 *-----------------------------------------------------------------------------*/
void harktool_getLocaltime(char *buff) {

    time_t timer;
    struct tm *local;

    /* 現在時刻を取得 */
    timer = time(NULL);

    local = localtime(&timer);

    /* 変換後表示 */
    sprintf(buff, "%4d%02d%02d%02d%02d%02d", local->tm_year + 1900, local->tm_mon + 1, local->tm_mday, local->tm_hour,
            local->tm_min, local->tm_sec);

} //end

/*-----------------------------------------------------------------------------*
 *
 * /tmp配下に使用するディレクトリを作成する
 *
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_createTmpPath() {

    int cnt;
    char path_name[FILENAME_MAX];

/*    if (harktool_CheckDir("/tmp") == NON_DIR_TYPE) {
        harktool_LogE("None [/tmp] Directory");
        harktool_Exit(EXIT_FAILURE);
    }*/

    char* tempdir = ".harktool-temp";
    if (harktool_CheckDir(tempdir) == NON_DIR_TYPE) {
        if (harktool_createDir(tempdir) == FAILURE) {
            harktool_LogE("Cannot create directory");
            harktool_Exit(EXIT_FAILURE);
        }
    }

    cnt = 0;
    while (cnt <= 5) { // 5回試みる
        harktool_getLocaltime(g.tmp.date_time);   //作成日時

        memset(path_name, '\0', sizeof(path_name));
        sprintf(path_name, "%s/%s", tempdir, g.tmp.date_time);

        if (harktool_CheckDir(path_name) == NON_DIR_TYPE) {
            if (harktool_createDir(path_name) == SUCCESS) {
                break;
            }
        }

        cnt++;
        sleep(1);
        printf("<< retry create directory >>\n");
    }

    //Directory check
    if (harktool_CheckDir(path_name) == NON_DIR_TYPE) {
        harktool_LogE("can not create directory");
        harktool_Exit(EXIT_FAILURE);
    }

    //set Directory Name
    sprintf(g.tmp.tsp_path, "%s/tsp", path_name);

    if (harktool_CheckDir(g.tmp.tsp_path) == NON_DIR_TYPE) {
        harktool_createDir(g.tmp.tsp_path);
    }

    sprintf(g.tmp.ir_path, "%s/imp", path_name);

    if (harktool_CheckDir(g.tmp.ir_path) == NON_DIR_TYPE) {
        harktool_createDir(g.tmp.ir_path);
    }

    sprintf(g.tmp.tf_path, "%s/tf", path_name);

    if (harktool_CheckDir(g.tmp.tf_path) == NON_DIR_TYPE) {
        harktool_createDir(g.tmp.tf_path);
    }

}   //end

/*-----------------------------------------------------------------------------*
 * ディレクトリの有無を確認する
 *
 *  0:directory -1: non directory
 *
 *
 *-----------------------------------------------------------------------------*/
int harktool_CheckDir(char *dirname)
//int harktool_CheckDir( char *dirname )
{

    struct stat sb;
    int ret = NON_DIR_TYPE;

    if (stat(dirname, &sb) == 0) {
        switch (sb.st_mode & S_IFMT) {
        case S_IFDIR: //0040000 ディレクトリ
            ret = DIR_TYPE;
            break;

        default:
            ret = NON_DIR_TYPE;
            break;
        }
    }

    return (ret);

} //end

/*-----------------------------------------------------------------------------*
 * ディレクトリ を作成する
 *
 *
 *
 *-----------------------------------------------------------------------------*/
int harktool_createDir(char* path) {

#ifndef _WIN32
    if (mkdir(path, 0755) == 0) {  //0:成功 -1:error
        return (SUCCESS);
    }
#else
    if (mkdir(path) == 0) {  //0:成功 -1:error
        return (SUCCESS);
    }
#endif
    return (FAILURE);

}

/*-----------------------------------------------------------------------------*
 * file sizeを取得する
 *
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_getFileSize(char* filename, long *size) {

    FILE *fp;

    fp = fopen(filename, "rb");

    if (fp == NULL) {
        char msg[MSG_LEN];
        sprintf(msg, "can not open file [%s]", filename);
        harktool_LogE(msg);
        harktool_Exit(EXIT_FAILURE);
    }

    fseek(fp, 0, SEEK_END);
    *size = ftell(fp);

    //printf( "%sファイルのサイズ:%d バイト", fname, sz );

    fclose(fp);

}  //end

/*-----------------------------------------------------------------------------*
 * 1ファイルのデータをbufferに格納する
 *
 *
 * Argument:
 *
 *  OUT : float *fltBuf // 格納先buffer
 *  IN  : long size     // ファイルサイズ
 *  IN  * char *fname   // 対象ファイル名
 *
 *-----------------------------------------------------------------------------*/
void harktool_SetFloatArray(float *fltBuf, long file_size, char *fname) {
    FILE* fp = fopen(fname, "rb");
    if (fp == NULL) {
        char msg[MSG_LEN];
        sprintf(msg, "can not open file [%s]", fname);
        harktool_LogE(msg);
        harktool_Exit(EXIT_FAILURE);
    }

    float size = file_size / sizeof(float);

    for (int i = 0; i < size; i++) {
        float read_buf;
        size_t sz = fread((void*) &read_buf, sizeof(float), 1, fp);
        if (sz == 1) {
            fltBuf[i] = read_buf;
        } else {
            harktool_LogE("fread error");
            fclose(fp);
            harktool_Exit(EXIT_FAILURE);
        }
    }

    fclose(fp);
}

/*-----------------------------------------------------------------------------*
 * 音源(SRC)ごとにインパルス応答を正規化する.
 *
 * in  : param.IRvector
 * out : param.tmp_IRvector
 *-----------------------------------------------------------------------------*/
void harktool_normalizeIR(float*** f3dvect) {
    // tmp領域にcopy
    harktool_copyTmpIRvectToTMP(f3dvect);

    //全データ(param.tmp_IRvector)を２乗
    harktool_powIR(g.TmpIRvect);

    //srcごとに総数を取得して、２乗した各データで割る
    for (int src = 0; src < g.Nsrc; src++) {
        // g.tmp_IRvect_vectorの総数を取得
        float sum = harktool_sumAll(src, normalize_src, g.TmpIRvect);

        // 多次元配列のすべての要素を指定した数値で割る
        // IRvector = tmp_IRvector/sum
        harktool_divAll(src, sum, normalize_src, f3dvect);
    }
}

/*-----------------------------------------------------------------------------*
 * 伝達関数を正規化する.
 *
 *
 *
 *-----------------------------------------------------------------------------*/
/*void harktool_normalizeTF()
 {


 // vector.TmpTFvectを初期化
 harktool_initArray(TMP_TF_VECT);


 //TFの最後の要素を"0"クリアする
 harktool_setZeroToTFvect();


 // vector.TFvect を vector.temp_tf にcopy
 harktool_copyTmpTFvectFromTFvect();



 //全データを絶対値にし、vector.TmpTFvectにセット
 harktool_absTF();

 //全データを2乗する(g.TmpTFvect)
 harktool_powTF();



 //各正規化処理
 if (g.argv.normalize_src == 1) {
 harktool_normalizeSRC();
 }

 if (g.argv.normalize_mic == 1) {
 harktool_normalizeMIC();
 }


 if (g.argv.normalize_freq == 1) {
 harktool_normalizeFREQ();
 }



 }//end*/

void normalizeTF(int normSrc, int normMic, int normFreq) {
    // vector.TmpTFvectを初期化
    harktool_initArray(TMP_TF_VECT);

    //TFの最後の要素を"0"クリアする
    harktool_setZeroToTFvect();

    // vector.TFvect を vector.temp_tf にcopy
    harktool_copyTmpTFvectFromTFvect();

    //全データを絶対値にし、vector.TmpTFvectにセット
    harktool_absTF();

    //全データを2乗する(g.TmpTFvect)
    harktool_powTF();

    //各正規化処理
    if (normSrc == 1) {
        harktool_normalizeSRC();
    }

    if (normMic == 1) {
        harktool_normalizeMIC();
    }

    if (normFreq == 1) {
        harktool_normalizeFREQ();
    }
}

/*-----------------------------------------------------------------------------*
 * 音源(SRC)ごとに伝達関数を正規化する.
 *
 *
 * in  : TmpTFvector
 * out : tf_vector
 *
 *-----------------------------------------------------------------------------*/
void harktool_normalizeSRC() {
    for (int mic = 0; mic < g.Nmic; mic++) {
        for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
            float sum = 0;

            for (int src = 0; src < g.Nsrc; src++) {
                //crealf 関数は，float complex 型の複素数 z の実部を float 型で返します
                sum += crealf(g.TmpTFvect[src][mic][fft]);
            }

            float sqrt_sum = sqrtf(sum);

            if (isfinite(sqrt_sum) != 0 && sqrt_sum != 0) {
                for (int src = 0; src < g.Nsrc; src++) {
                    float real = crealf(g.TFvect[src][mic][fft]) / sqrt_sum; //normで割る
                    float imag = cimagf(g.TFvect[src][mic][fft]) / sqrt_sum; //normで割る

                    g.TFvect[src][mic][fft] = real + imag * _Complex_I;
                }
            }
        }
    }

}

/*-----------------------------------------------------------------------------*
 * @brief MICごとに伝達関数を正規化する.
 *
 * sqrtf 関数は引数 x の平方根を計算し，結果を float 型で返します．
 *
 * in  : TmpTFvector
 * out : tf_vector
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_normalizeMIC() {
    for (int src = 0; src < g.Nsrc; src++) {
        for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
            float sum = 0;

            for (int mic = 0; mic < g.Nmic; mic++) {
                //crealf 関数は，float complex 型の複素数 z の実部を float 型で返します
                sum = sum + crealf(g.TmpTFvect[src][mic][fft]);
            }

            float sqrt_sum = sqrtf(sum);

            if (isfinite(sqrt_sum) != 0 && sqrt_sum != 0) {
                for (int mic = 0; mic < g.Nmic; mic++) {
                    float real = crealf(g.TFvect[src][mic][fft]) / sqrt_sum; //normで割る
                    float imag = cimagf(g.TFvect[src][mic][fft]) / sqrt_sum; //normで割る

                    g.TFvect[src][mic][fft] = real + imag * _Complex_I;
                }
            }
        }
    }
}

/*-----------------------------------------------------------------------------*
 * @brief 周波数(FRQ)ごとに伝達関数を正規化する.
 *
 *
 *
 * in  : TmpTFvector
 * out : tf_vector
 *
 *-----------------------------------------------------------------------------*/
void harktool_normalizeFREQ() {
    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {

            float sum = 0;

            for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
                //crealf 関数は，float complex 型の複素数 z の実部を float 型で返します
                sum = sum + crealf(g.TmpTFvect[src][mic][fft]);
            }

            float sqrt_sum = sqrtf(sum);

            if (isfinite(sqrt_sum) != 0 && sqrt_sum != 0) {
                for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
                    float real = crealf(g.TFvect[src][mic][fft]) / sqrt_sum; //normで割る
                    float imag = cimagf(g.TFvect[src][mic][fft]) / sqrt_sum; //normで割る

                    g.TFvect[src][mic][fft] = real + imag * _Complex_I;

                }
            }
        }
    }
}

void showReverbSources(int src, int peakNearestIdx, const int* peaks) {
    bool showReverbsource = false;
    for (int mic = 0; mic < g.Nmic; mic++) {
        if (DIRECT_SOUND_POSITION_LIMIT <= (peaks[mic] - peakNearestIdx)) {
            if (!showReverbsource) {
                printf("\n");
                printf("     ----------------------------------------------------------------------------\n");
                printf("            Direct sound < Reverberant sound                                     \n");
                printf("     ----------------------------------------------------------------------------\n");
                showReverbsource = true;
            }

            printf("[ %s ]  Mic:[%d]ch  Max Peak Index[%d]\n",
                    harkio_Positions_getById(g.src_xml->poses, g.SrcPosId[src])->path, g.MicPosId[mic], peaks[mic]);
            cat_asprintf(&g.readmeMsg, "Direct sound < Reverberant sound  [ %s ]  Mic:[%d]ch  Max Peak Index[%d]\n",
                    harkio_Positions_getById(g.src_xml->poses, g.SrcPosId[src])->path, g.MicPosId[mic], peaks[mic]);
        }
    }
}

/*-----------------------------------------------------------------------------*
 * fft 実行
 *
 * 1. vector.buff_a <- IRvector
 * 2. TFvect <- fft()
 *
 * c = std::complex< double >( 3.0, 4.0 );
 *
 *
 * in  : vector.IRvect
 * out : vector.TFvect
 *
 *-----------------------------------------------------------------------------*/
void harktool_FFT(float*** f3dvect) {
    memset(g.buff_A, '\0', g.Nfft * sizeof(COMPLEX_FLOAT));

    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {

            // fft実行 a[]の値をfftした結果がA[]にセットされる
            harktool_FFTImpl_fRealFFT(f3dvect[src][mic], g.buff_A, g.Nfft);

            //Nfft/2 +1分コピー
            for (int fft = 0; fft < g.Nfft / 2 + 1; fft++) {
                g.TFvect[src][mic][fft] = crealf(g.buff_A[fft]) + cimagf(g.buff_A[fft]) * _Complex_I;
            }
        }
    }
}

/*-----------------------------------------------------------
 * 伝達関数をconjする
 *
 *
 *  conjf 関数は，複素数 z の共役複素数を float complex 型で返します
 *
 *
 * in  : g.TFvect
 * out : g.TFvect
 *
 -----------------------------------------------------------*/
void harktool_conjTF() {
    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
                g.TFvect[src][mic][fft] = conjf(g.TFvect[src][mic][fft]);
            }
        }
    }
}

int searchPeak(const float* irVect, int irLength, int start, int end) {
    int peak;
    float max;
    int i;
    if (start < 0 || end < 0) {
        // all range
        peak = 0;
        max = fabsf(irVect[0]);
        i = 1;
        end = irLength;
    } else {
        // [start - end] range
        if (start > irLength) {
            start = irLength;
        }
        peak = start;
        max = fabsf(irVect[start]);
        i = start + 1;
        if (end > irLength) {
            end = irLength;
        }
    }
    // search
    for (; i < end; i++) {
        if (max < fabsf(irVect[i])) {
            max = fabsf(irVect[i]);
            peak = i;
        }
    }
    return peak;
}

/*-----------------------------------------------------------------------------*
 *  g.TFvect のデータを、g.TmpTFvect にcopyする
 *
 * in  : g.TFvect
 * out : g.TmpTFvect
 *-----------------------------------------------------------------------------*/
void harktool_copyTmpTFvectFromTFvect() {
    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
                g.TmpTFvect[src][mic][fft] = crealf(
                        g.TFvect[src][mic][fft]) + cimagf(g.TFvect[src][mic][fft]) * _Complex_I;
            }
        }
    }
}

/*-----------------------------------------------------------------------------*
 * g.IRvect のデータを、g.tmp_IRvectにcopyする
 *
 * in  : g.IRvect
 * out : g.TmpIRvect
 *-----------------------------------------------------------------------------*/
void harktool_copyTmpIRvectToTMP(float*** f3dvect) {
    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < g.Nfft; fft++) {
                g.TmpIRvect[src][mic][fft] = f3dvect[src][mic][fft];
            }
        }
    }
}

/*-----------------------------------------------------------------------------*
 * tf のデータを絶対値し、tmp_tfにセットする
 *
 * in  : g.TmpTFvect
 * out : g.TmpTFvect
 *-----------------------------------------------------------------------------*/
void harktool_absTF() {
    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
                //cabsf 関数は，float complex 型の引数 z の絶対値を float 型で返します
                g.TmpTFvect[src][mic][fft] = cabsf(g.TmpTFvect[src][mic][fft]);
            }
        }
    }
}

/*-----------------------------------------------------------------------------*
 * tmp_tf の全値を2 乗する
 *
 * x の 2 乗を計算し，結果を float 型で返します
 *
 * in   : g.TmpTFvect
 * out  : g.TmpTFvect
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_powTF() {
    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
                g.TmpTFvect[src][mic][fft] *= g.TmpTFvect[src][mic][fft];

                //cpowf 関数は，x の y 乗を計算し，結果を float complex 型で返します．
                //g.TmpTFvect[src][mic][fft] =  crealf(value) + cimagf(value) * _Complex_I;
            }
        }
    }
}

/*-----------------------------------------------------------------------------*
 * ir の全値を2 乗する
 *
 * x の 2 乗を計算し，結果を float 型で返します
 *
 * in  : IRvector[][][]
 * out : IRvector[][][]
 *
 *-----------------------------------------------------------------------------*/
void harktool_powIR(float*** f3dvect) {
    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < g.Nfft; fft++) {
                f3dvect[src][mic][fft] = powf(f3dvect[src][mic][fft], 2.0);
            }
        }
    }
}

/*-----------------------------------------------------------------------------*
 * @brief Impulse応答の総和を求める.（SRCごと）
 *
 *
 *  in  : IRvector[][][]
 *  out : float sum
 *
 *-----------------------------------------------------------------------------*/
float harktool_sumAll(int src, int type, float*** f3dvect) {
    float sum = 0;

    switch (type) {
    case normalize_src:
        for (int mic = 0; mic < g.Nmic; mic++) {
            float sumMic = 0;
            for (int fft = 0; fft < g.Nfft; fft++) {
                sumMic += f3dvect[src][mic][fft];
            }
            sum += sumMic;
        }

        break;

//  case normalize_mic:
//  case normalize_freq:

    }

    return sum;
}

/*-----------------------------------------------------------------------------*
 * Impulse応答の要素を指定した数値で割る
 *
 *
 *   in  : int   src
 *       : float sum
 *       : int   type
 *       : IRvector[][][]
 *   out : IRvector[][][]
 *
 *-----------------------------------------------------------------------------*/
void harktool_divAll(int src, float sum, int type, float*** f3dvect) {
    switch (type) {
    case normalize_src:
        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < g.Nfft; fft++) {
                f3dvect[src][mic][fft] /= sum;
            }
        }
        break;

    default:
        break;
    }
}

/*-----------------------------------------------------------------------------*
 * 伝達関数をZipファイルに出力する
 *
 *
 *
 *-----------------------------------------------------------------------------*/
/*void harktool_writeTF()
 {

 int ret;



 //harkio_TransferFunction型にセット

 switch (g.PosType) {
 case TSP:
 g.TF = harkio_TransferFunction_fromData(g.tsp_xml->poses
 , g.mic_xml->poses
 , g.tsp_xml->cfg
 , g.tsp_xml->nbrs);
 break;

 case IMPULSE:
 g.TF = harkio_TransferFunction_fromData(g.src_xml->poses
 , g.mic_xml->poses
 , g.src_xml->cfg
 , g.src_xml->nbrs);
 break;

 default:
 harktool_LogE("harkio_TransferFunction_write error");
 harktool_Exit(EXIT_FAILURE);
 }


 //伝達関数をセット
 switch (g.argv.tf_type) {
 case LOC:// LOCALIZATION
 harkio_TransferFunction_setLocTFs(g.TF, g.Loctf);
 break;

 case SEP: // SEPARATION
 harkio_TransferFunction_setSepTFs(g.TF, g.Septf);
 break;

 case LOC_SEP:
 harkio_TransferFunction_setLocTFs(g.TF, g.Loctf);
 harkio_TransferFunction_setSepTFs(g.TF, g.Septf);
 break;
 }


 ret = harkio_TransferFunction_write(g.TF , g.argv.output , NULL);

 if (ret == EXIT_FAILURE) {
 harktool_LogE("harkio_TransferFunction_write error");
 harktool_Exit(EXIT_FAILURE);
 }

 }//end*/

void writeTF(enum TFType tfType, const char* output, bool resetMic) {
    int ret;

    if (resetMic) {
        // reest microphone info
        int nMic = g.mic_xml->poses->size;
        for (int i = 0; i < nMic; i++) {
            bool useChannel = false;
            for (int j = 0; j < g.Nmic; j++) {
                if (i == g.MicPosId[j]) {
                    useChannel = true;
                    break;
                }
            }
            if (!useChannel) {
                harkio_Positions_removeById(g.mic_xml->poses, i);
            }
        }
        nMic = g.mic_xml->poses->size;
        for (int i = 0; i < nMic; i++) {
            g.mic_xml->poses->pos[i]->id = i;
        }
        // remove <channels use> from src
        for (int i = 0; i < g.src_xml->poses->size; i++) {
            harkio_Position_setChannelsUse(g.src_xml->poses->pos[i], 0);
        }
    }

    //harkio_TransferFunction型にセット
    g.TF = harkio_TransferFunction_fromData(g.src_xml->poses, g.mic_xml->poses, g.src_xml->cfg, g.src_xml->nbrs);

    //伝達関数をセット
    switch (tfType) {
    case LOC:                // LOCALIZATION
        harkio_TransferFunction_setLocTFs(g.TF, g.Loctf);
        break;

    case SEP: // SEPARATION
        harkio_TransferFunction_setSepTFs(g.TF, g.Septf);
        break;

    case LOC_SEP:
        harkio_TransferFunction_setLocTFs(g.TF, g.Loctf);
        harkio_TransferFunction_setSepTFs(g.TF, g.Septf);
        break;
    }

    ret = harkio_TransferFunction_write(g.TF, output, NULL);

    if (ret == EXIT_FAILURE) {
        harktool_LogE("harkio_TransferFunction_write error");
        harktool_Exit(EXIT_FAILURE);
    }
}

/*-----------------------------------------------------------------------------*
 * 定位伝達関数の配列からharkio_Matrix型に変換する
 *
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_setLoctfFromArray() {
    int len = g.Nmic * (g.Nfft / 2 + 1);

    //Nsrc個 harkio_matrix を作成
    g.Loctf = (harkio_Matrix**) malloc(sizeof(harkio_Matrix*) * g.Nsrc);

    if (g.Loctf == NULL) {
        harktool_LogE("malloc Failed");
        harktool_Exit(EXIT_FAILURE);
    }

    for (int src = 0; src < g.Nsrc; src++) {
        g.Loctf[src] = harkio_Matrix_new("complex", g.Nmic, (g.Nfft / 2 + 1));

        if (g.Loctf[src] == NULL) {
            harktool_LogE("harkio_Matrix_new Failed");
            harktool_Exit(EXIT_FAILURE);
        }

        memset(g.Loctf[src]->data.complexdata, '\0', len * sizeof(COMPLEX_TYPE));

        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
                harkio_Matrix_setValueComplex(g.Loctf[src], mic, fft, g.TFvect[src][mic][fft]);
            }
        }
    }
}

/*-----------------------------------------------------------------------------*
 * 分離伝達関数の配列からharkio_Matrix型に変換する
 *
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_setSeptfFromArray() {

    int len = g.Nmic * (g.Nfft / 2 + 1);

    //Nsrc個 harkio_matrix を作成
    g.Septf = (harkio_Matrix**) malloc(sizeof(harkio_Matrix*) * g.Nsrc);

    if (g.Septf == NULL) {
        harktool_LogE("malloc Failed");
        harktool_Exit(EXIT_FAILURE);
    }

    for (int src = 0; src < g.Nsrc; src++) {
        g.Septf[src] = harkio_Matrix_new("complex", g.Nmic, (g.Nfft / 2 + 1));
//        g.Septf[src] = harkio_Matrix_new( "complex" , 1 , len );//1 source bun

        if (g.Septf[src] == NULL) {
            harktool_LogE("harkio_Matrix_new Failed");
            harktool_Exit(EXIT_FAILURE);
        }

        memset(g.Septf[src]->data.complexdata, '\0', len * sizeof(COMPLEX_TYPE));

        for (int mic = 0; mic < g.Nmic; mic++) {
            for (int fft = 0; fft < (g.Nfft / 2 + 1); fft++) {
                harkio_Matrix_setValueComplex(g.Septf[src], mic, fft, g.TFvect[src][mic][fft]);
//                g.Septf[src]->data.complexdata[mic] = crealf( g.TFvect[src][mic][fft] ) + cimagf( g.TFvect[src][mic][fft] ) * _Complex_I;
            }
        }
    }
}

/*-----------------------------------------------------------------------------*
 * 伝達関数の最後の要素を0にする
 *
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_setZeroToTFvect() {
    float real = 0.0f;
    float imag = 0.0f;

    int fft = g.Nfft / 2; //最後の値

    //最後の値に"0"をセット
    for (int src = 0; src < g.Nsrc; src++) {
        for (int mic = 0; mic < g.Nmic; mic++) {
            g.TFvect[src][mic][fft] = real + imag * _Complex_I;
        }
    }
}

/*-----------------------------------------------------------------------------*
 * 終了処理
 *-----------------------------------------------------------------------------*/
void harktool_freeMemory() {

    /*
     harktool_freeComplexThreeDim( g.TFvect , g.Nsrc , g.Nmic  );
     harktool_freeComplexThreeDim( g.TmpTFvect , g.Nsrc , g.Nmic  );
     harktool_freeFloatThreeDim( g.IRvect , g.Nsrc , g.Nmic  );
     harktool_freeFloatThreeDim( g.TmpIRvect, g.Nsrc , g.Nmic  );
     harktool_freeFloatTwoDim( g.wav.data , g.Nsrc );
     harktool_freeFloatTwoDim( g.MicPosVect, g.Nmic );
     harktool_freeCharTwoDim( g.tmp.flt_path , g.Nsrc  );
     harktool_freeCharTwoDim( g.tmp.tsp_wav_name , g.Nsrc  );
     //  harktool_freeCharTwoDim( g.tmp.ir_wav_path , g.Nsrc  );
     harktool_freeIntOneDim( g.MicPosId  );
     harktool_freeIntOneDim( g.IrPosId   );
     harktool_freeComplexOneDim( g.buff_a  );
     harktool_freeComplexOneDim( g.buff_A  );
     harktool_freeComplexOneDim( g.buff_w  );
     */

    harktool_delete(TF_VECT);
    harktool_delete(TMP_TF_VECT);
    harktool_delete(IR_VECT);
    harktool_delete(TMP_IR_VECT);
    harktool_delete(WAV_DATA);
    harktool_delete(MIC_POS_VECT);
    harktool_delete(FLT_PATH);
    harktool_delete(TSP_WAV_NAME);
    harktool_delete(SORT_POS_ID);
    harktool_delete(BUFFER);

    if (g.mic_xml != NULL)
        free(g.mic_xml);  // 伝達関数作成時のマイク位置
    if (g.src_xml != NULL)
        free(g.src_xml);  // 伝達関数作成時の音源位置
    if (g.noise_xml != NULL)
        free(g.noise_xml);  // 伝達関数作成時のノイズ位置

    //TF delete
    harkio_TransferFunction_delete(&g.TF);

    if (g.readmeMsg)
        free(g.readmeMsg);
}

// Idで指定されている座標をposVectに格納する
// 座標(position->sys)が極座標の場合は、直交座標に変換する
int setPos(const harkio_Positions* poses, const int* posIds, int nId, float** posVect) {
    for (int i = 0; i < nId; i++) {
        int id = posIds[i];

        //idのpositionを取得
        harkio_Position* pos = harkio_Positions_getById(poses, id);

        if (pos->sys == Polar) {
            float a_coord[3];
            if (harkio_Position_Convert(pos, a_coord, Cartesian) == EXIT_FAILURE) {
                harkio_Log_printf(harkio_Log_Error, "harkio_Position_Convert error");
                return EXIT_FAILURE;
            }
            posVect[i][0] = a_coord[0]; // x
            posVect[i][1] = a_coord[1]; // y
            posVect[i][2] = a_coord[2]; // z
        } else {
            posVect[i][0] = pos->coord[0]; // x
            posVect[i][1] = pos->coord[1]; // y
            posVect[i][2] = pos->coord[2]; // z
        }
    }
    return EXIT_SUCCESS;
}

void harktool_setMicPos() {
    if (setPos(g.mic_xml->poses, g.MicPosId, g.Nmic, g.MicPosVect) == EXIT_FAILURE) {
        harktool_Exit(EXIT_FAILURE);
    }
}

void harktool_setSrcPos() {
    if (setPos(g.src_xml->poses, g.SrcPosId, g.Nsrc, g.SrcPosVect) == EXIT_FAILURE) {
        harktool_Exit(EXIT_FAILURE);
    }
}

/*-----------------------------------------------------------------------------*
 * mic positionの平均
 *-----------------------------------------------------------------------------*/
//void harktool_calcMicPosMean() {
//    for (int i = 0; i < 3; i++) {
//        g.PosMean[i] = 0;
//        for (int j = 0; j < g.Nmic; j++) {
//            g.PosMean[i] += g.MicPosVect[j][i];
//        }
//        g.PosMean[i] /= g.Nmic;
//    }
//}
// MIC ID 設定
void harktool_setMicPosId(const bool* outChannels) {
    int mic = 0;
    for (int i = 0; i < MAX_CH; i++) {
        if (outChannels[i]) {
            g.MicPosId[mic] = i;
            mic++;
        }
        if (mic >= g.Nmic)
            break;
    }
}

void harktool_setSrcPosId() {
    for (int src = 0; src < g.Nsrc; src++) {
        g.SrcPosId[src] = g.src_xml->poses->pos[src]->id;
    }

    // sort
    harktool_BubSort(g.SrcPosId, g.Nsrc);
}

/*-----------------------------------------------------------------------------*
 * ソート処理
 *-----------------------------------------------------------------------------*/
/* バブルソートを行う */
void harktool_BubSort(int x[], int n) {
    int i, j, temp;

    for (i = 0; i < n - 1; i++) {
        for (j = n - 1; j > i; j--) {
            if (x[j - 1] > x[j]) { /* 前の要素の方が大きかったら */
                temp = x[j]; /* 交換する */
                x[j] = x[j - 1];
                x[j - 1] = temp;
            }
        }
    }

    harktool_ShowData(x, n);
}

/* ソート対象のデータを表示 */
void harktool_ShowData(int x[], int n) {
    for (int i = 0; i < n; i++) {
        harkio_Log_printf(harkio_Log_Debug, "x[%d]=%d ", i, x[i]);
    }
}

/*----------------------------------------------------------*/
/*                                                          */
/*----------------------------------------------------------*/
long get_flt_size(const char* fname) {
    if (fname == NULL) {
        return -1;
    }

    FILE* fd = fopen(fname, "rb");

    if (fd == NULL) {
        return -1;
    }

    fseek(fd, 0, SEEK_END);
    long file_size = ftell(fd);

    fclose(fd);

    return file_size;
}

/*----------------------------------------------------------*/
/*                                                          */
/*----------------------------------------------------------*/
int read_flt(const char* fname, double* dBuf, size_t file_size, double* signal_max) {
    if (fname == NULL || dBuf == NULL || signal_max == NULL || (file_size % sizeof(float)) != 0) {
        return -1;
    }

    FILE* fd = fopen(fname, "rb");

    if (fd == NULL) {
        return -1;
    }

    char file_buf[BUFSIZ];
    setvbuf(fd, file_buf, _IOFBF, BUFSIZ); // full buffering

    int result = 0;
    float read_buf;
    *signal_max = 0;
    size_t i;

    for (i = 0; i < file_size / sizeof(float); ++i) {
        size_t r = fread((void*) &read_buf, sizeof(float), 1, fd);
        if (r < 1) {
            // error
            result = -1;
            break;
        }
        dBuf[i] = (double) read_buf;
        const double amp = fabs(dBuf[i]);
        if (amp > *signal_max) {
            *signal_max = amp;
        }
    }

    fclose(fd);
    return result;
}

/*----------------------------------------------------------*/
/*                                                          */
/*----------------------------------------------------------*/
int write_flt(const char* fname, const double* dBuf, size_t count) {
    if (fname == NULL || dBuf == NULL) {
        return -1;
    }

    FILE* fd = fopen(fname, "wb");

    if (fd == NULL) {
        return -1;
    }

    char file_buf[BUFSIZ];
    setvbuf(fd, file_buf, _IOFBF, BUFSIZ); // full buffering

    int result = 0;
    float write_buf;
    int i;

    for (i = 0; i < count; ++i) {
        write_buf = (float) dBuf[i];

        size_t r = fwrite((void*) &write_buf, sizeof(float), 1, fd);
        if (r < 1) {
            // error
            result = -1;
            break;
        }
    }

    fflush(fd);
    fclose(fd);
    return result;
}

// read wave data
int readwave(SNDFILE *fp, SF_INFO *sfinfo, int count, double ***wave) {

    if ((*wave = (double **) calloc(sfinfo->channels, sizeof(double *))) == NULL) {
        perror("calloc");
        return -1;
    }

    //assert(count <= sfinfo->frames);
    for (int i = 0; i < sfinfo->channels; i++) {
        if (((*wave)[i] = (double *) calloc(count, sizeof(double))) == NULL) {
            perror("calloc");
            free(*wave);
            return -1;
        }
    }

    // read normalized [-1.0, 1.0] wave data as double
    double *data;

    if ((data = (double *) calloc(sfinfo->channels, sizeof(double))) == NULL) {
        perror("calloc");
        for (int i = 0; i < sfinfo->channels; i++) {
            free((*wave)[i]);
        }
        free(*wave);
        return -1;
    }

    for (int i = 0; i < count; i++) {
        if (sf_readf_double(fp, data, 1) != 1) {
            printf("cannot read frame=%d\n", i);
            break;
        }
        for (int j = 0; j < sfinfo->channels; j++) {
            (*wave)[j][i] = data[j];
        }
    }

    free(data);

    return 1;
}

// read wave file
int readwavfile(const char *fname, double ***wave, int *length, int *ch, int *samplerate) {
    SNDFILE *fp;
    SF_INFO sfinfo;

    sfinfo.format = 0;
    if ((fp = sf_open(fname, SFM_READ, &sfinfo)) == NULL) {
        return -1;
    }
    int ret = readwave(fp, &sfinfo, sfinfo.frames, wave);
    sf_close(fp);
    *length = sfinfo.frames;
    *ch = sfinfo.channels;
    *samplerate = sfinfo.samplerate;

    return ret;
}

/*----------------------------------------------------------*/
/*                                                          */
/*----------------------------------------------------------*/
// write wave data
int writewave(SNDFILE *fp, SF_INFO *sfinfo, int count, double **wave) {

    short *data;
    int i, j;

    if ((data = (short *) malloc(sizeof(short) * count * sfinfo->channels)) == NULL) {
        perror("malloc");
        exit(0);
    }

    for (i = 0; i < count; i++)
        for (j = 0; j < sfinfo->channels; j++) {
            data[j + i * sfinfo->channels] = (short) wave[j][i];
        }

    sf_writef_short(fp, data, count);

    free(data);

    return 1;
}

/*----------------------------------------------------------*/
/*                                                          */
/*----------------------------------------------------------*/
// read wave file
int writewavfile(char *fname, double **wave, int length, int ch, int samplerate) {
    SNDFILE *fp;
    SF_INFO sfinfo;

    sfinfo.format = SF_FORMAT_WAV | SF_FORMAT_PCM_16;
    sfinfo.samplerate = samplerate;
    sfinfo.channels = ch;
    sfinfo.sections = -1;
    sfinfo.seekable = -1;
    //  sfinfo.frames = length;

    fp = sf_open(fname, SFM_WRITE, &sfinfo);
    writewave(fp, &sfinfo, length, wave);
    sf_close(fp);

    return 1;
}

int appendInfoToZipFile(const char *zip_path, const SRC_POS_INFO* srcPosInfo) {
    if (zip_path == NULL || g.readmeMsg == NULL)
        return EXIT_SUCCESS;

    int errorp;
    struct zip *s;
    struct zip_source *src;

    //file open
    s = zip_open(zip_path, ZIP_CREATE, &errorp);

    if (s == NULL) {
        char errmsg[100];
        zip_error_to_str(errmsg, sizeof(errmsg), errorp, errno);
        harkio_Log_printf(harkio_Log_Error, "zip open error errorp[%d] (%s)", errorp, errmsg);
        return EXIT_FAILURE;
    }

    // create README
    src = zip_source_buffer(s, g.readmeMsg, strlen(g.readmeMsg), 0);

    if (src == NULL) {
        harkio_Log_printf(harkio_Log_Error, "zip_source_buffer error (%s)", zip_strerror(s));
        return EXIT_FAILURE;
    }

    zip_add(s, "transferFunction/README", src);

    // append position info to source.xml
    char* srcPath = "transferFunction/source.xml";
    char* strSrcXml = NULL;
    if (as_zip_fread(s, srcPath, &strSrcXml) != EXIT_SUCCESS) {
        zip_close(s);
        return EXIT_FAILURE;
    }
    // parse and modify source.xml
    xmlInitParser();
    xmlDocPtr doc = xmlParseDoc(BAD_CAST strSrcXml);
    free(strSrcXml);
    if (doc == NULL) {
        harkio_Log_printf(harkio_Log_Error, "source.xml parse error");
        zip_close(s);
        return EXIT_FAILURE;
    }
    xmlXPathContextPtr xpathCtx = xmlXPathNewContext(doc);
    if (xpathCtx == NULL) {
        harkio_Log_printf(harkio_Log_Error, "xmlXPathNewContext error");
        xmlFreeDoc(doc);
        zip_close(s);
        return EXIT_FAILURE;
    }
    //xmlNodePtr comment = xmlNewComment(BAD_CAST "add TFTimeOffset and CutIndex by HARKTOOL5");
    for (int i = 0; i < g.Nsrc; i++) {
        if (srcPosInfo[i].loc.tfOffset == 0.0 && srcPosInfo[i].loc.cutFrom == 0 && srcPosInfo[i].sep.tfOffset == 0.0
                && srcPosInfo[i].sep.cutFrom == 0)
            continue;
        char xpathExpr[100];
        sprintf(xpathExpr, "/hark_xml/positions/position[@id=%d]", g.SrcPosId[i]);
        xmlXPathObjectPtr xpathObj = xmlXPathEvalExpression(BAD_CAST xpathExpr, xpathCtx);
        if (xpathObj == NULL) {
            harkio_Log_printf(harkio_Log_Error, "xmlXPathEvalExpression error");
            xmlXPathFreeContext(xpathCtx);
            xmlFreeDoc(doc);
            zip_close(s);
            return EXIT_FAILURE;
        }
        //printf("position size=%d\n", xpathObj->nodesetval->nodeNr);
        if (xpathObj->nodesetval->nodeNr > 0) {
            xmlNodePtr posNode = xpathObj->nodesetval->nodeTab[0];
            //printf("path=%s\n", (char*)xmlGetProp(posNode, BAD_CAST "path"));
            //xmlAddPrevSibling(posNode->parent, comment);
            char strOffset[20];
            for (enum TFType type = LOC; type <= SEP; type++) {
                const POSITION_INFO* posInfo = (type == LOC) ? &(srcPosInfo[i].loc) : &(srcPosInfo[i].sep);
                if (posInfo->tfOffset == 0.0 && posInfo->cutFrom == 0)
                    continue;
                sprintf(strOffset, "%f", posInfo->tfOffset);
                xmlNodePtr node = xmlNewChild(posNode, NULL, BAD_CAST "calcTF",
                NULL);
                xmlSetProp(node, BAD_CAST "type",
                BAD_CAST ((type == LOC) ? "localization" : "separation"));
                xmlSetProp(node, BAD_CAST "offset", BAD_CAST strOffset);
                if (posInfo->cutFrom > 0 && posInfo->cutTo > 0) {
                    char strFrom[10], strTo[10];
                    sprintf(strFrom, "%d", posInfo->cutFrom);
                    sprintf(strTo, "%d", posInfo->cutTo);
                    xmlNewProp(node, BAD_CAST "cutIndexFrom", BAD_CAST strFrom);
                    xmlNewProp(node, BAD_CAST "cutIndexTo", BAD_CAST strTo);
                }
            }
        }
        xmlXPathFreeObject(xpathObj);
    }
    xmlXPathFreeContext(xpathCtx);
    xmlChar* xmlBuff;
    int xmlBuffSize;
    xmlDocDumpMemory(doc, &xmlBuff, &xmlBuffSize);
    char* outBuff = calloc(xmlBuffSize, sizeof(char));
    strncpy(outBuff, (char*) xmlBuff, xmlBuffSize);
    xmlFree(xmlBuff);
    xmlFreeDoc(doc);
    xmlCleanupParser();
    // replace source.xml
    src = zip_source_buffer(s, outBuff, xmlBuffSize, 0);
    if (src == NULL) {
        harkio_Log_printf(harkio_Log_Error, "zip_source_buffer error (%s)", zip_strerror(s));
        zip_close(s);
        return EXIT_FAILURE;
    }
    if (zip_replace(s, zip_name_locate(s, srcPath, 0), src) == -1) {
        harkio_Log_printf(harkio_Log_Error, "zip_replace error (%s)", zip_strerror(s));
        zip_close(s);
        return EXIT_FAILURE;
    }

    zip_close(s);

    return EXIT_SUCCESS;
}

/*-----------------------------------------------------------------------------*
 * 終了処理
 *
 *
 *
 *-----------------------------------------------------------------------------*/
void harktool_Exit(int exit_type) {

    if (exit_type == EXIT_SUCCESS) {
        harktool_LogI("harktool normal end");
        harktool_freeMemory();
        exit(EXIT_SUCCESS);
    }

    harktool_LogE("harktool abnormal end");

    harktool_freeMemory();

    exit(EXIT_FAILURE);

}   //end

/*void harktool_genMUSIC()
 {
 // wavファイルからIR変換
 harktool_setIRvectFromWav();

 // 残響音と思われる部分に0を書き込む
 harktool_EraseReverberation();

 // 残響音が直接音より大きい可能性がある音源を表示
 harktool_ShowReverbSources();

 // インパルス応答をFFTして伝達関数を取得
 harktool_FFT();

 //定位伝達関数の配列からharkio_Matrix型に変換する
 harktool_setLoctfFromArray();

 }*/

void genMUSIC() {
    // インパルス応答をFFTして伝達関数を取得
    harktool_FFT(g.LocIRvect);

    //定位伝達関数の配列からharkio_Matrix型に変換する
    harktool_setLoctfFromArray();
}

/*void harktool_genGSS()
 {

 // インパルス応答の正規化
 harktool_normalizeIR();


 // インパルス応答をFFTして伝達関数を取得
 harktool_FFT();

 // conj
 harktool_conjTF();

 harktool_normalizeTF();

 //分離伝達関数の配列からharkio_Matrix型に変換する
 harktool_setSeptfFromArray();

 }*/

void genGSS(int normSrc, int normMic, int normFreq) {
    /* インパルス応答の正規化 */
    harktool_normalizeIR(g.SepIRvect);

    // インパルス応答をFFTして伝達関数を取得
    harktool_FFT(g.SepIRvect);

    /* conj */
    harktool_conjTF();

    normalizeTF(normSrc, normMic, normFreq);

    // write normalization setting to README
    cat_asprintf(&g.readmeMsg, "%s \n%s[%d]\n%s[%d]\n%s[%d]\n", "Normalization Type", "- Source     ", normSrc,
            "- Microphone ", normMic, "- Frequency  ", normFreq);

    //分離伝達関数の配列からharkio_Matrix型に変換する
    harktool_setSeptfFromArray();
}

void genTfFromIR(enum TFType outputType, int normSrc, int normMic, int normFreq,
bool resetMic, const char* outputFile, const SRC_POS_INFO* srcPosInfo) {
    switch (outputType) {
    //outputType:SEP,LOC,LOC_SEP
    case LOC:
        genMUSIC();
        break;
    case SEP:
        genGSS(normSrc, normMic, normFreq);
        break;
    case LOC_SEP:
        //Loc
        genMUSIC();
        //data free
        harktool_delete(TMP_IR_VECT);
        harktool_delete(TF_VECT);
        harktool_delete(TMP_TF_VECT);
        harktool_delete(BUFFER);
        harktool_new(TMP_IR_VECT); // impulse応答 計算用領域
        harktool_new(TF_VECT); // 伝達関数用領域
        harktool_new(TMP_TF_VECT); // 伝達関数 作業用領域
        harktool_new(BUFFER); // 伝達関数作成用TEMP領域
        //sep
        genGSS(normSrc, normMic, normFreq);
        break;
    }
    /*--------------------------------------------------*/
    // 伝達関数をZipファイルに出力
    /*--------------------------------------------------*/
    writeTF(outputType, outputFile, resetMic);
    /*--------------------------------------------------*/
    // READMEに正規化情報を書き込む
    /*--------------------------------------------------*/
    appendInfoToZipFile(outputFile, srcPosInfo);
}

void makeTempFltPath(int src, const harkio_Position* pos) {
    char* str;
    str = strrchr(pos->path, '/');
    char* strPath = NULL;
    if (str != NULL) {
        //wav名を格納
        sprintf(g.tmp.tsp_wav_name[src], "%s", str + 1);
        // /tmp/harktool/日時/tsp/xxx.wav ディレクトリ作成
        asprintf(&strPath, "%s/%s", g.tmp.tsp_path, str + 1);
        if (harktool_CheckDir(strPath) == NON_DIR_TYPE) {
            harktool_createDir(strPath);
        }
    } else {
        harkio_Log_printf(harkio_Log_Error, "can not get path [%s] ", pos->path);
        harktool_Exit(EXIT_FAILURE);
    }
    //path copy
    sprintf(g.tmp.flt_path[src], "%s", strPath); //fltファイル格納先path
    if (strPath)
        free(strPath);

    /// outimpfilehead = "/tmp/harktool/xxx.wav/float"
    sprintf(g.tmp.outimpfilehead, "%s/%s", g.tmp.flt_path[src], FILE_PREFIX);
}
