/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

﻿#if 0
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fltfile.h"

int get_flt_size(char *fname) {
    int file_size;

    if (fname == NULL) {
        return -1;
    }

    FILE *fd = fopen(fname, "rb");

    if ((fd < 0) || (fd == NULL)) {
        return -1;
    }

    fseek(fd, 0, SEEK_END);
    file_size = ftell(fd);

    fclose(fd);

    return file_size;
}

int read_flt(char *fname, double* dBuf, int size) {
    char *file_buf;
    float read_buf;
    int idx = 0;
    if (fname == NULL) {
        return -1;
    }

    FILE *fd = fopen(fname, "rb");

    if ((fd < 0) || (fd == NULL)) {
        return -1;
    }

    file_buf = (char *) malloc(size);
    setbuf(fd, file_buf);

    for (idx = 0; idx > -1; idx++) {
        if (fread((void *) &read_buf, sizeof(float), 1, fd)) {
            dBuf[idx] = read_buf;
        } else {
            break;
        }
    }

    fflush(fd);

    fclose(fd);

    free(file_buf);

    return 0;
}

int read_flt(char *fname, double* dBuf, int size, double* signal_max) {
    char *file_buf;
    float read_buf;
    int idx = 0;
    if (fname == NULL) {
        return -1;
    }

    FILE *fd = fopen(fname, "rb");

    if ((fd < 0) || (fd == NULL)) {
        return -1;
    }

    *signal_max = 0;
    file_buf = (char *) malloc(size);
    setbuf(fd, file_buf);

    for (idx = 0; idx > -1; idx++) {
        if (fread((void *) &read_buf, sizeof(float), 1, fd)) {
            dBuf[idx] = read_buf;
            if(dBuf[idx] > *signal_max){
                *signal_max = fabs(dBuf[idx]);
            }
        } else {
            break;
        }
    }

    fflush(fd);

    fclose(fd);

    free(file_buf);

    return 0;
}


int write_flt(char *fname, double* dBuf, int count) {
    char *file_buf;
    float write_buf;
    int idx = 0;
    if (fname == NULL) {
        return -1;
    }

    FILE *fd = fopen(fname, "wb");

    if (fd < 0) {
        return -1;
    }

    file_buf = (char *) malloc(count * sizeof(float));
    setbuf(fd, file_buf);

    for (idx = 0; idx < count; idx++) {
        write_buf = dBuf[idx];
        fwrite((void *) &write_buf, sizeof(float), 1, fd);
    }

    fflush(fd);

    fclose(fd);

    free(file_buf);

    return 0;
}

/*int main()
 {
 return 0;
 }*/

#else
//#include <cstdio>
//#include <cmath>
#include<stdio.h>
#include<stdlib.h>
#include "harktool_fltfile.h"

long get_flt_size(const char* fname)
{
    if (fname == NULL)
    {
        return -1;
    }

    FILE* fd = fopen(fname, "rb");

    if (fd == NULL)
    {
        return -1;
    }

    fseek(fd, 0, SEEK_END);
    long file_size = ftell(fd);

    fclose(fd);

    return file_size;
}

#if 0
int read_flt(const char* fname, double* dBuf, size_t file_size)
{
    double dummy;
    return read_flt(fname, dBuf, file_size, &dummy);
}
#endif

int read_flt(const char* fname, double* dBuf, size_t file_size, double* signal_max)
{
    if (fname == NULL || dBuf == NULL || signal_max == NULL ||
            (file_size % sizeof(float)) != 0)
    {
        return -1;
    }

    FILE* fd = fopen(fname, "rb");

    if (fd == NULL)
    {
        return -1;
    }

    char file_buf[BUFSIZ];
    setvbuf(fd, file_buf, _IOFBF, BUFSIZ); // full buffering

    int result = 0;
    float read_buf;
    *signal_max = 0;
    size_t i;

    for (  i = 0; i < file_size / sizeof(float); ++i)
    {
        size_t r = fread((void*)&read_buf, sizeof(float), 1, fd);
        if (r < 1)
        {
            // error
            result = -1;
            break;
        }
        dBuf[i] = (double) read_buf;
        const double amp = fabs(dBuf[i]);
        if (amp > *signal_max)
        {
            *signal_max = amp;
        }
    }

    fclose(fd);
    return result;
}

int write_flt(const char* fname, const double* dBuf, size_t count)
{
    if (fname == NULL || dBuf == NULL)
    {
        return -1;
    }

    printf("@@@ write_flt start fname[%s] count[%u]\n",fname ,count);

    FILE* fd = fopen(fname, "wb");

    if (fd == NULL)
    {
        return -1;
    }

    char file_buf[BUFSIZ];
    setvbuf(fd, file_buf, _IOFBF, BUFSIZ); // full buffering

    int result = 0;
    float write_buf;
    size_t i;

    for ( i = 0; i < count; ++i)
    {
        write_buf = (float)dBuf[i];

        if(i<100)
        {
            //printf("[%d]write_flt[%20f] [%x] double[%f]\n",i,write_buf,  *(int*)&write_buf ,dBuf[i])	;
        }

        size_t r = fwrite((void*)&write_buf, sizeof(float), 1, fd);
        if (r < 1)
        {
            // error
            result = -1;
            break;
        }
    }

    fflush(fd);
    fclose(fd);
    return result;
}


#endif

