/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "harktool.h"

//global 構造体
struct GLOBAL g;

bool hasParsedOpt = false;

void help() {
    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  HARKTOOL\n");
    printf("\n");
    printf("Description:\n");
    printf("  HARKTOOL is a tool for visualization and generation of the transfer function\n");
    printf("\n");
    printf("Function List:\n");
    printf("  1. calctfimp\n");
    printf("       Generate transfer function from impulse responses\n");
    printf("          Usage : harktool5 calctfimp --help\n");
    printf("\n");
    printf("  2. calctfrec\n");
    printf("       Generate transfer function from TSPs\n");
    printf("          Usage : harktool5 calctfrec --help\n");
    printf("\n");
    printf("  3. calctfgeo\n");
    printf("       Generate transfer function by geometric calculation\n");
    printf("          Usage : harktool5 calctfgeo --help\n");
    printf("\n");
    printf("  4. calctfmouth\n");
    printf("       Generate transfer function from mouth TSPs\n");
    printf("          Usage : harktool5 calcsysimp --help\n");
    printf("\n");
    printf("  5. calcsysimp\n");
    printf("       calculate impluse response from TSP wave\n");
    printf("          Usage : harktool5 calcsysimp --help\n");
    printf("\n");
    printf("\n");
}

/*-----------------------------------------------------------------------------*
 * main処理
 *-----------------------------------------------------------------------------*/
int main(int argc, char *argv[]) {
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error, "Main Argument Error");
        harkio_Log_printf(harkio_Log_Error,
                "\n\nHow to use [harktool5 --help]\n\n");
        exit(EXIT_FAILURE);
    }

    /*--------------------------------------------------*/
    /*   help                                           */
    /*--------------------------------------------------*/
    if (strcmp(argv[1], "--help") == 0) {
        help();
        exit(EXIT_SUCCESS);
    }

    if (argc == 2) {
        harkio_Log_printf(harkio_Log_Error, "Main Argument Error");
        harkio_Log_printf(harkio_Log_Error,
                "\n\nHow to use [harktool5 --help]\n\n");
        exit(EXIT_FAILURE);
    }

    // インパルス応答から伝達関数生成
    int ret;
    if (strcmp(argv[1], "calctfimp") == 0) {
        ret = harktoolcli_calctfimp(argc, argv);
    }
    // TSP録音から伝達関数作成処理
    else if (strcmp(argv[1], "calctfrec") == 0) {
        ret = harktoolcli_calctfrec(argc, argv);
    }
    // 幾何計算から伝達関数作成処理
    else if (strcmp(argv[1], "calctfgeo") == 0) {
        ret = harktoolcli_calctfgeo(argc, argv);
    }
    // くちTSP録音から伝達関数作成処理
    else if (strcmp(argv[1], "calctfmouth") == 0) {
        ret = harktoolcli_calctfmouth(argc, argv);
    }
    // TSP録音からインパルス応答(flt)を生成
    else if (strcmp(argv[1], "calcsysimp") == 0) {
        ret = harktoolcli_calcsysimp(argc, argv);
    } else {
        harkio_Log_printf(harkio_Log_Error, "Main Argument Error. Unknown function [%s]", argv[1]);
        exit(EXIT_FAILURE);
    }

    exit(ret);
}

