/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <fftw.h>
#include <rfftw.h>
#include <sndfile.h>
#include "readimp.h"
#include "wavfile.h"
#include "floatfile.h"
#include "myconv.h"




extern void harktool_writeWavFile( char *fname, float *buffer , int frames , int nmic , int samplerate );

#if 0
void usage()
{
	printf("./conv inputfile impfilehead nch [noisefile alpha] outfile\n");
	printf("compute : out = alpha x imp * input + noise\n");
	printf("inputfile   : 16bit short mono wave format\n");
	printf("impfilehead : its extention is k.flt where k is ch number.\n");
	printf("              each imp file is 32bit float mono 1-ch raw data\n");
	printf("  eg.) when impfilehead = 'a', file names are 'a1.flt', 'a2.flt', ...\n");
	printf("nch         : num of channels = num of imp files\n");
	printf("alpha       : gain (default is 1.0)\n");
	printf("noisefile   : 16bit short x-ch wave format (x >= n)\n");
	printf("outfile     : 32bit float LE n-ch raw data\n");
		
	exit(1);
	/*./myconv ../../../org/f101b001.wav ../../../../impulse/imp2wav/050.wav 2 ./outtmp.wav*/

}
#endif


// for debug out
void usage()
{
	printf("                                                                                    \n");
	printf("                                                                                    \n");
	printf("[NAME:]                                                                             \n");
	printf("  harktoolcli-conv                                                                  \n");
	printf("                                                                                    \n");
	printf("[Description:]                                                                      \n");
	printf("   convolution tool                                                                 \n");
	printf("                                                                                    \n");
	printf("[Usage :]                                                                           \n");
    printf("  harktoolcli-conv [Argument] [option]                                              \n");
	printf("                                                                                    \n");
    printf("[Argument :]                                                                        \n");
    printf("  -i :  Impulse response file           [ float 32bit wave format]                  \n");
    printf("  -c :  input  file (before convolution)[ signed 16bit PCM wave format: short mono] \n");
    printf("  -o :  output file (after convolution) [wav file]                                  \n");
	printf("                                                                                    \n");
    printf("  [option]                                                                          \n");
    printf("  -s :   select channels of output [ex. -s 0,3,6]                                   \n");
	printf("                                                                                    \n");
    exit(EXIT_FAILURE);
}













int main(int argc, char *argv[])
{

    int     i, j;
    float  *outwavf;          // output data
  //  int     wavlen;
    double *wavres;           // tmp
    int     ch;               // num of channels
//	char *sname = NULL;       // signal file name
//	char *ihead = NULL;       // imp file name (offset)
//	char *oname = NULL;       // output file name
    char   *nname = NULL;     // noise file name
    double  alpha = 1.0;      // gain
	
    char    infname[1024];    //input  file(clean wav)
    char    impfname[1024];   //Impulse response file
    char    outfname[1024];   //output file
    int     outchannel[1024]; //select channel of output
    int     outchcount = 0;   //select channel count



    memset(infname    ,'\0' , sizeof(infname) );
    memset(impfname   ,'\0' , sizeof(impfname) );
    memset(outfname   ,'\0' , sizeof(outfname) );
    memset(outchannel ,'\0' , sizeof(outchannel) );

    for(i=0;i<argc;i++)
    {
        printf("%s ",argv[i]);
    }

    printf("\n");


    if( argc <= 4 )
    {
        usage();
    }


#if 0

	if( argc != 5 && argc != 7 ){
		usage();
	}

    sname = argv[1];  //inputfile
    ihead = argv[2];  //impfilehead
    ch=atoi(argv[3]); //nch
	if( ch < 1 ){
		printf("error: num of ch\n");
		exit(1);
	}

	if( argc == 5 ){
        oname = argv[4]; //outfile
		nname = NULL;
		alpha = 1.0;
	} else if(argc == 7){
        nname = argv[4];       //noisefile
        alpha = atof(argv[5]); //alpha
        oname = argv[6];       //outfile
		if( alpha <= 0.0 ){
			fprintf(stderr, "error: alpha val %lf\n", alpha);
			exit(1);
		}
	}
#endif

    int option;

    // parse option
    while(( option = getopt(argc, argv, "i:c:o:s:")) != -1)
    {
        switch(option){
            case 'i': //impulse-file
                strcpy(impfname , optarg);  //impulse-file
                break;

            case 'c': //input-file
                strcpy(infname , optarg);  //inputfile
                break;

            case 'o': //output_file
                strcpy(outfname , optarg);  //inputfile
                break;

            case 's':
                outchcount = split( optarg ,  outchannel );
                break;

            case '?':
            usage();
        }
    }









    //--------------------------------------------------------//
    // read signal file
    // 16bit short mono wave format
    //--------------------------------------------------------//
    double **swave;
    int    slength;
    int    sch;
    int    ssamplerate;
    int    sformat    ; // format

    printf("read signal %s...\n", infname);

    readwavfile(infname, &swave, &slength, &sch, &ssamplerate ,&sformat);




    //--------------------------------------------------------//
	// read impulse response file
    //--------------------------------------------------------//
//	char   impfname[1024]; // imp file name
    double **impdata;     // impulse response data
    int    implength;     // length of impulse response data
    int    impch;         // channel num
    int    impsamplerate; // samplerate
    int    impformat    ; // format


    printf("read impulse %s...\n", impfname);

//    readwavfile(impfname, &impdata, &implength, &impch, &impsamplerate ,&impformat);




    readwavimpfile(impfname, &impdata, &implength, &impch, &impsamplerate ,&impformat);



    ch = impch;


    printf( "read impulse response -> # of ch: %d\n", ch );


#if 0
    if( (impdata = (double **)malloc(sizeof(double *) * ch)) == NULL ){
		perror("malloc");
		exit(1);
	}

	printf( "read impulse response -> # of ch: %d\n", ch );

    for( i = 0; i < ch; i++ )
    {
		sprintf(impfname, "%s%d.flt",ihead, i+1);
		//		printf("read %s...", impfname);
		readimp(impfname, &(impdata[i]), &implength);
		//		printf("%d bytes\n", implength);
	}



	//
	// read noise file
	//
	double **nwave;
    int    nlength;
    int    nch;
    int    nsamplerate;
    int    nstart; // startoffset;

    if( nname != NULL )
    {
		printf("read noise %s...\n", nname);
		readwavfile(nname, &nwave, &nlength, &nch, &nsamplerate);
	}

    if( nname != NULL && nlength < slength )
    {
		fprintf(stderr, "noise length is too short.\n");
	}
#endif




//    nstart = (int)(((float)rand()/RAND_MAX) * (nlength - slength));

    outwavf = (float *)malloc(sizeof(float) * slength * ch); // 4byte * slength(frame) * channel



    //---------------------------------------------------//
	// computation
    //---------------------------------------------------//

	double maxval = 0.0;
	int wavlen=0 ;
	
    for(i=0;i<ch;i++)
    {
		//
		// compute convolution
		//
        wavlen = convolusion(swave[0], slength, impdata[i], implength, &wavres);

		//
		// compute X = aHS + N
		//
        for(j=0; j<slength; j++)
        {
            if( nname != NULL ) //noise file
            {
//                outwavf[j*ch+i] = (float)(wavres[j] * alpha ) + nwave[i][nstart+j];
            }
			else
            {
                outwavf[j*ch+i] = (float)(wavres[j] * alpha );
                printf("wavres[%d]=%f\n",j,wavres[j]);
            }

			if( maxval < fabs(outwavf[j*ch+i]) )
            {
				maxval = fabs(outwavf[j*ch+i]);
            }
		}

        free(wavres);
	}


    //---------------------------------------------------//
    // write output file
    //---------------------------------------------------//
    printf("write output file %s...\n", outfname);

//    writefloatfile(oname, outwavf, slength, ch);




    float *data;

    if( outchcount != 0)
    {

        //sizeof(float) * frame(slength) * outchcount
        if( ( data = ( float *)malloc( sizeof(float) * slength * outchcount) ) == NULL )
        {
            perror("malloc");
            exit(0);
        }


        //-s のチャンネルのみ取得(slength=frame)
        int cunt = 0;
        for( j = 0; j <  slength; j++ )
        {
            for( i = 0; i < ch; i++ )
            {
                if( i == outchannel[i])
                {
                    //                         [ fram * Nch + i]
                    data[cunt] = (float)outwavf[ j * ch  + i ] ;
                    cunt++;
                }
            }
        }


         harktool_writewavfile(outfname    ,
                               data        ,
                               slength     ,
                               outchcount  ,
                               ssamplerate );

         free(data);
    }
    // all channel
    else
    {
        harktool_writewavfile(outfname    ,
                              outwavf        ,
                              slength     ,
                              impch  ,
                              ssamplerate );


    }






    free(outwavf);

	printf("max amplitude = %lf\n", maxval);
	
    exit(EXIT_SUCCESS) ;

}








int split( char *s1 ,  int *intCh )
{
    char s2[] = ", ";
    char *tok;
    int  cnt = 0;

    tok = strtok( s1, s2 );
    intCh[cnt] = atoi(tok);
    cnt++;

    while( tok != NULL )
    {
        intCh[cnt] = atoi(tok);
        cnt++;
        tok = strtok( NULL, s2 );  /* 2回目以降 */
    }

    return (cnt+1);
}




